/* comm/speech.c */
/* Commands which involve speaking */

#include "externs.h"

dbref speaker=NOTHING;

/* track @redirect command */
dbref redirect_plr=NOTHING;
dbref redirect_rcpt=NOTHING;

/* track @iterate command */
dbref iterate_player=NOTHING;
char *iterate_pattern;
char *iterate_command;


/* check for spoofing in a room */
static int can_emit_msg(player, target, loc, msg)
dbref player, target, loc;
char *msg;
{
  char mybuf[8200], *s;
  dbref thing, first;

  strcpy(mybuf, msg);
  if((s=strchr(mybuf,' ')))
    *s=0;
  if((thing=lookup_player(mybuf)) > NOTHING && thing != target &&
     !strcmp(db[thing].name, mybuf) && !controls(player, thing, POW_SPOOF))
    return 0;
  if(Invalid(loc))
    return 1;
  DOLIST(first, db[loc].contents)
    if(first != target && !strcmp(db[first].name, mybuf) &&
       !controls(player, first, POW_SPOOF))
      return 0;
  return 1;
}

/* relay message. bypass puppet and check for ! events */
static void notify_relay(dbref player, char *msg)
{
  dbref *ptr;
  static int remote=0, depth=0;
  char *d;

  /* check if string matches @listen wildcards */
  if(db[speaker].location != NOTHING && *(d=atr_get(player, A_LISTEN)) &&
     wild_match(msg, d, 0, 1)) {
    /* activate ahear on object */
    did_it(speaker,player,0,NULL,0,NULL,A_AHEAR);

    /* activate @relay messages to other room */
    if(Typeof(player) != TYPE_ROOM && db[speaker].location != player)
      notify_except(player, player, msg);
    if(!remote && *(d=atr_get(player, A_RHEAR)) && db[player].relay) {
      char buf2[strlen(d)+1], buf3[8192];

      strcpy(buf2, d);
      pronoun_substitute(buf3, player, speaker, buf2);

      remote=1;
      for(ptr=db[player].relay;ptr && *ptr != NOTHING;ptr++)
        if(power(player, POW_SPOOF) || can_emit_msg(player,NOTHING,*ptr,buf3))
          notify_except(*ptr, NOTHING, buf3);
      remote=0;
    }
  }

  /* now check for multi listeners */
  if(++depth < 3 && speaker != player && could_doit(speaker, player, A_ULOCK))
    atr_match(player, speaker, '!', msg);
  depth--;
}

/* send message to user */
void notify(dbref player, char *msg)
{
  if(Invalid(player))
    return;
  if(player == redirect_plr)
    player=redirect_rcpt;

  /* redirect output through @iterate command */
  if(speaker == player && player == iterate_player) {
    if(wild_match(msg, iterate_pattern, 0, 1))
      parse_que(player, iterate_command, player);
    return;
  }

  /* messages to players go direct to network output */
  if(Typeof(player) == TYPE_PLAYER)
    raw_notify(player, msg);
  else {
    /* if player is a puppet, relay message to puppet's owner */
    if((db[player].flags & PUPPET) && IS(db[player].owner, TYPE_PLAYER,
        PLAYER_CONNECT)) {
      char buf[strlen(db[player].name)+strlen(msg)+3];

      sprintf(buf, "%s> %s", db[player].name, msg);
      raw_notify(db[player].owner, buf);
    }
    notify_relay(player, msg);
  }
}

/* regular speech command */
void do_say(player, arg1, arg2, argv, pure)
dbref player;
char *arg1, *arg2, **argv, *pure;
{
  char *s;

  if(!pass_slock(player, 1))
    return;

  s=parse_function(player, player, pure);

  notify(player, tprintf("You say, \"%s\"", s));
  notify_except(db[player].location, player, tprintf("%s says \"%s\"",
                db[player].name, s));
}

/* whisper aloud to someone in the room */
void do_whisper(dbref player, char *arg1, char *arg2)
{
  char *s;
  dbref who, first;
  
  if(Typeof(player) != TYPE_PLAYER) {
    notify(player, "Only players may whisper.");
    return;
  }

  if((who=lookup_player(arg1)) <= NOTHING) {
    notify(player, tprintf("No such player '%s'.", arg1));
    return;
  } if(player == who) {
    notify(player, "That's rather hard.");
    return;
  } if(!power(player,POW_REMOTE) && (db[who].location != db[player].location ||
       !(db[who].flags & PLAYER_CONNECT))) {
    notify(player, "That player isn't here.");
    return;
  }

  s=parse_function(player, player, arg2);

  if(pass_slock(player, 0))
    DOLIST(first, db[db[player].location].contents) {
      if(first == player || first == who)
        continue;
      notify(first, tprintf(
             "You notice %s whisper something in %s%s ear.",
             db[player].name, db[who].name, possname(db[who].name)));
    }

  if(*s == ':' || *s == ';')  {
    notify(player, tprintf("You whisper-posed %s with \"%s%s %s\".",
           db[who].name, db[player].name, *s==';'?possname(db[player].name):
           "", s+1));
    notify(who, tprintf("%s whisper-poses: %s%s %s",
           db[player].name, db[player].name, *s==';'?
           possname(db[player].name):"", s+1));
  } else {
    notify(player, tprintf("You whisper \"%s\" to %s.", s, db[who].name));
    notify(who, tprintf("%s whispers \"%s\" to you.", db[player].name, s));
  }
}

/* pose a message */
void do_pose_cmd(player, arg1, arg2, argv, pure)
dbref player;
char *arg1, *arg2, **argv, *pure;
{
  do_pose(player, pure, 0);
}

void do_pose(dbref player, char *msg, int type)
{
  char *s;
  
  if(!pass_slock(player, 1))
    return;

  s=parse_function(player, player, msg);

  notify_except(db[player].location, NOTHING, tprintf("%s%s %s",
                db[player].name, type?possname(db[player].name):"", s));
}


/* general emote commands */

enum { EMIT, PEMIT, REMIT, OEMIT, OREMIT, ZEMIT, PRINT,
       NOPARSE=0x80 };

void do_echo(player, arg1)
dbref player;
char *arg1;
{
  notify(player, parse_function(player, player, arg1));
}

void do_necho(player, arg1, arg2, argv, pure)
dbref player;
char *arg1, *arg2, **argv, *pure;
{
  notify(player, pure);
}

void do_emit(player, arg1)
dbref player;
char *arg1;
{
  room_emote(player, NOTHING, 0, arg1, EMIT);
}

void do_nemit(player, arg1, arg2, argv, pure)
dbref player;
char *arg1, *arg2, **argv, *pure;
{
  room_emote(player, NOTHING, 0, pure, EMIT|NOPARSE);
}

void do_pemit(player, arg1, arg2)
dbref player;
char *arg1, *arg2;
{
  room_emote(player, NOTHING, arg1, arg2, PEMIT);
}

void do_npemit(player, arg1, arg2, argv, pure)
dbref player;
char *arg1, *arg2, **argv, *pure;
{
  room_emote(player, NOTHING, arg1, pure, PEMIT|NOPARSE);
}

void do_remit(player, arg1, arg2)
dbref player;
char *arg1, *arg2;
{
  room_emote(player, NOTHING, arg1, arg2, REMIT);
}

void do_oemit(player, arg1, arg2)
dbref player;
char *arg1, *arg2;
{
  room_emote(player, NOTHING, arg1, arg2, OEMIT);
}

void do_oremit(player, arg1, arg2)
dbref player;
char *arg1, *arg2;
{
  room_emote(player, NOTHING, arg1, arg2, OREMIT);
}

void do_zemit(player, arg1, arg2)
dbref player;
char *arg1, *arg2;
{
  room_emote(player, NOTHING, arg1, arg2, ZEMIT);
}

void do_print(player, arg1, arg2, argv, pure, cause)
dbref player, cause;
char *arg1, *arg2, **argv, *pure;
{
  room_emote(player, cause, 0, arg1, PRINT);
}

void room_emote(player, thing, target, msg, type)
dbref player, thing;
char *target, *msg;
int type;
{
  dbref i, loc=db[player].location;
  char buff[8192], *s=buff;
  
  /* parse message */
  if(type & NOPARSE) {
    type &= ~NOPARSE;
    s=msg;
    if(type == PEMIT)
      while(*s)
        if(*s++ == '=')
          break;
  } else
    strcpy(buff, parse_function(player, player, msg));

  /* check target */
  if(thing == NOTHING && target) {
    if((type == PEMIT || type == OEMIT) && !power(player, POW_REMOTE))
      thing=match_thing(player, target, MATCH_CON|MATCH_INVEN|MATCH_ME|
                        MATCH_HERE|MATCH_INZONE);
    else
      thing=match_thing(player, target, MATCH_EVERYTHING);
    if(thing == NOTHING)
      return;

    switch(type) {
    case PEMIT:
      loc=db[thing].location;
      break;
    case REMIT:
      if(thing != db[player].location && !controls(player, thing, POW_REMOTE)) {
        notify(player, "Permission denied.");
        return;
      }
      loc=thing;
      thing=NOTHING;
      break;
    case OREMIT:
      if(db[thing].location != db[player].location && !controls(player, thing,
         POW_REMOTE)) {
        notify(player, "Permission denied.");
        return;
      }
      loc=db[thing].location;
      break;
    case ZEMIT:
      if(!controls(player, thing, POW_MODIFY) || !can_emit_msg(player,
         NOTHING, NOTHING, s)) {
        notify(player, "Permission denied.");
        return;
      } if(Typeof(thing) != TYPE_ZONE) {
        notify(player, "That is not a zone object.");
        return;
      }
      loc=NOTHING;
    }
  } else if(type == PRINT && Typeof(player) != TYPE_ZONE &&
            mainroom(player) != mainroom(thing))
    return;

  if(loc == db[player].location)
    if(!pass_slock(player, 1))
      return;

  if(loc != NOTHING && !controls(player, thing, POW_SPOOF) &&
     !can_emit_msg(player, thing, loc, s)) {
    notify(player, "Permission denied.");
    return;
  }
  
  /* display the message */

  if(type == PEMIT || type == PRINT)
    notify(thing, s);
  else if(type == ZEMIT) {
    for(i=0;i<db_top;i++)
      if(inlist(db[i].zone, thing))
        notify_except(i, NOTHING, s);
  } else
    notify_except(loc, thing, s);
}

/* can you talk in a room? */
int pass_slock(dbref player, int type)
{
  if(power(player,POW_SPOOF) || could_doit(player,db[player].location,A_SLOCK))
    return 1;

  if(type)
    ansi_did_it(player, db[player].location, A_SFAIL, "The room is muted.",
                0, 0, A_ASFAIL);
  return 0;
}

void do_announce(player, arg1, argc, argv, arg2)
dbref player;
char *arg1, *argc, **argv, *arg2;
{
  char buf[9000];
  DESC *d;

  /* check permissions */
  if(!*arg2) {
    notify(player, "No message.");
    return;
  } if(Typeof(player) != TYPE_PLAYER && !power(player, POW_SPOOF)) {
    notify(player, "You can't do that.");
    return;
  }

  /* guests cannot announce */
  if(Guest(player)) {
    notify(player, "Guests cannot announce messages. Use the [public] channel "
           "instead, by typing \"=your message\" on a blank line (without the "
           "quotes).");
    return;
  }

  /* only users who've played more than 2 hours can announce messages */
  if(ANNOUNCE_COST && Typeof(player) == TYPE_PLAYER &&
     db[player].data->age < 7200) {
    notify(player, tprintf(
         "This command announces your message to everyone online and costs "
         "%d %s. That probably isn't what you want right now. Try using the "
         "[public] channel instead, by typing \"=your message\" on a blank "
         "line (without the quotes).", ANNOUNCE_COST, MONEY_PLURAL));
    return;
  }
  
  /* check costs */
  if(ANNOUNCE_COST && !power(player, POW_ANNOUNCE)) {
    if(!charge_money(player, ANNOUNCE_COST, 0)) {
      notify(player, tprintf("It takes %d %s to make an announcement.",
             ANNOUNCE_COST, MONEY_PLURAL));
      return;
    }
    notify(player,tprintf("You have been charged %d %s for that announcement.",
           ANNOUNCE_COST, MONEY_PLURAL));
  }

  if(*arg2 == ':' || *arg2 == ';')
    sprintf(buf, "%s announce-poses: %s%s %s\n", db[player].name,
            db[player].name, (*arg2==':')?"":possname(db[player].name),arg2+1);
  else
    sprintf(buf, "%s announces: %s\n", db[player].name, arg2);

  for(d=descriptor_list;d;d=d->next)
    if(d->flags & C_CONNECTED)
      output(d, buf);
}

void do_broadcast(player, arg1, argc, argv, arg2)
dbref player;
char *arg1, *argc, **argv, *arg2;
{
  DESC *d;
  
  if(!*arg2) {
    notify(player, "No message.");
    return;
  }
  
  for(d=descriptor_list;d;d=d->next)
    if(!(d->flags & (C_OUTBOUND|C_SOCKET|C_DNS|C_PARENT|C_SERVICE)))
      output2(d, tprintf("Official broadcast from %s: %s\n", db[player].name,
              arg2));
}

/* wizard-emit, wall-emit, world-emit, whatever.. */
void do_wemit(player, arg1, argc, argv, arg2)
dbref player;
char *arg1, *argc, **argv, *arg2;
{
  DESC *d;
  
  for(d=descriptor_list;d;d=d->next)
    if(d->flags & C_CONNECTED)
      output(d, tprintf("%s\n", arg2));
}

/* high-priority output message to player's connection */
void do_output(dbref player, char *arg1, char *arg2)
{
  DESC *d;
  dbref thing=match_thing(player, arg1, MATCH_EVERYTHING);

  if(thing == NOTHING)
    return;

  if(Typeof(thing) != TYPE_PLAYER || !(d=match_desc(thing)))
    notify(player, "That player is not logged in.");
  else
    output2(d, tprintf("%s\n", arg2));
}

void do_page(dbref player, char *arg1, char *arg2)
{
  DESC *d=match_desc(player);
  char buf[512], mudname[32], *p, *r, *s;
  int i, j, hidden, count=0, suppress=0, remote=power(player, POW_REMOTE);
  dbref thing, obj[20];

  if(Typeof(player) != TYPE_PLAYER && !power(player, POW_SPOOF)) {
    notify(player, "Objects cannot page anyone.");
    return;
  }

  /* Retrieve playerlist from last page command if <arg1> is blank */
  if(!*arg1) {
    if(!d || !*d->env[3]) {
      notify(player, "You must specify a list of players.");
      notify(player, "Usage: page <player list>=<message>");
      return;
    }
    strmaxcpy(buf, d->env[3], 512);
    s=buf;
  } else {
    if(d)
      WPTR(d->env[3], arg1);
    s=arg1;
  }

  if(!*arg2) {
    notify(player, "No message specified.");
    return;
  }

  /* Store and count recipients */
  while((r=parse_up(&s, ' '))) {
    /* Recipients of the form User@Mudname use RWHO database */
    if((p=strchr(r, '@'))) {
      strmaxcpy(mudname, p+1, 28);
      if(strlen(mudname) < 4 || strcasecmp(mudname+strlen(mudname)-4, "MARE"))
        strcat(mudname, "MARE");
      if(strcasecmp(mudname, MUD_NAME)) {
        notify(player, "TinyMARE RWHO Server not yet supported.");
        continue;
      }
      *p='\0';
    }

    /* Local Server */
    if((obj[count]=lookup_player(r)) <= NOTHING) {
      notify(player, tprintf("No such player '%s'.", r));
      continue;
    }
    for(i=0;i<count && obj[i] != obj[count];i++);
    if(i == count && ++count > 20) {
      notify(player, "Too many players listed.");
      return;
    }
  }

  /* check for sufficient funds */
  if(!charge_money(player, PAGE_COST*count, 0)) {
    notify(player, tprintf("You don't have enough %s.", MONEY_PLURAL));
    return;
  }

  /* validate the list of players */
  for(i=0;i<count;i++) {
    thing=obj[i];
    if(player == thing)
      suppress=-1;

    /* Check paging permissions on target player */
    hidden=0;
    if(!(db[thing].flags & PLAYER_CONNECT) || (!controls(player, thing,
       POW_WHO) && !could_doit(player, thing, A_HLOCK))) {
      if(controls(player, thing, POW_WHO))
        notify(player, tprintf("%s is not connected.", db[thing].name));
      else
        notify(player,tprintf("%s is seemingly not connected.",
               db[thing].name));
      if(*(s=atr_parse(thing, player, A_AWAY)))
        notify(player, tprintf("Away message from %s: %s", db[thing].name, s));
      if(!(db[thing].flags & PLAYER_CONNECT))
        continue;
      hidden=1;
    }
    if(!could_doit(player, thing, A_PLOCK)) {
      if(!hidden) {
        notify(player,tprintf("%s%s is not contactable.",remote?"Warning: ":"",
               db[thing].name));
        if(*(s=atr_parse(thing, player, A_HAVEN)))
          notify(player,tprintf("Haven message from %s: %s",db[thing].name,s));
      }
      if(!remote)
        continue;
    }

    /* Check paging permissions on person doing the page */
    if(!could_doit(thing, player, A_PLOCK) && !remote) {
      notify(player, tprintf("Your pages are locked to %s.", db[thing].name));
      continue;
    }

    /* Build [Time], Name, and (Alias) Prefix */
    s=buf;
    if(*(atr_get(thing, A_IDLE)))
      s+=sprintf(s, "\e[1;36m[\e[37m%-8.8s\e[36m]\e[m ", mktm(thing, now)+11);
    s+=sprintf(s, "%s%s", textcolor(7, unparse_color(thing, player, 14)),
               db[player].name);
    if(*(r=atr_get(player, A_ALIAS)) && strlen(r) <= 10)
      s+=sprintf(s, " (%s)", r);
    s+=sprintf(s, " pages");

    /* Build Recipients list */
    if(count > 1) {
      for(j=0;j<count;j++) {
        if(j && j == count-1) {
          if(count > 2)
            *s++=',';
          strcpy(s, " and");
          s+=4;
        } else if(j)
          *s++=',';
        *s++=' ';
        s+=sprintf(s, "%s", (thing == obj[j])?"you":db[obj[j]].name);
      }
      strcpy(s, " with");
    }

    /* Special formatting for poses */
    j=(*arg2 == ':' || *arg2 == ';');
    notify(thing, tprintf("%s: %s%s%s%s\e[m", buf, j?db[player].name:"",
           (*arg2 == ';')? possname(db[player].name):"", j?" ":"", arg2+j));

    if(player != thing && *(s=atr_get(thing, A_APAGE))) {
      strcpy(wptr[0], arg2);
      parse_que(thing, s, player);
      *wptr[0]='\0';
    }

    if(hidden)
      continue;

    if(!suppress)
      suppress=1;
    if((j=getidletime(thing)) > 300 && *(s=atr_parse(thing, player, A_IDLE)))
      notify(player, tprintf("%s is %s idle: %s", db[thing].name,
             time_format(TMA, j), s));
  }

  if(suppress != 1)
    return;

  /* Build final names list for pager to view */
  for(s=buf,j=0;j<count;j++) {
    if(j && j == count-1) {
      if(count > 2)
        *s++=',';
      strcpy(s, " and ");
      s+=5;
    } else if(j) {
      *s++=',';
      *s++=' ';
    }
    s+=sprintf(s, "%s%s", db[obj[j]].name, isbusy(obj[j])?"(BUSY)":"");
  }

  j=(*arg2==':' || *arg2==';');
  notify(player, tprintf("You paged %s with: %s%s%s%s", buf,
         j?db[player].name:"", (*arg2==';')?possname(db[player].name):"",
         j?" ":"", arg2+j));
}

/* give message to all in room except.... */
void notify_room(dbref first, dbref exception, dbref exception2, char *msg)
{
  char buf[strlen(msg)+1];

  if(IS(first, TYPE_ROOM, ROOM_XZONE))
    return;

  /* strcpy to deal with notifying tprintf()'s */
  strcpy(buf, msg);
  if(first != exception && first != exception2)
    notify(first, msg);

  DOLIST(first, db[first].contents)
    if(first != exception && first != exception2) {
      if(IS(first, TYPE_THING, PUPPET) && db[first].location ==
         db[db[first].owner].location)
        notify_relay(first, msg);
      else
        notify(first, msg);
    }
}

/* internal procedure to display messages to both player and others in room */
/*  %N = player,  %U = cause.  */
void room_emit(player, cause, msg1, msg2)
dbref player, cause;
char *msg1, *msg2;
{
  if(msg1 && *msg1)
    notify(player, parse_function(cause, player, msg1));
  if(msg2 && *msg2)
    notify_except(db[player].location, player,
         parse_function(cause, player, msg2));
}

void do_cnotify(player, message, loc)
dbref player;
char *message;
int loc;
{
  DESC *d;
  char attr[8192], *s;

  for(d=descriptor_list;d;d=d->next) {
    if(!(d->flags & C_CONNECTED) || d->player == player || (loc &&
       db[player].location == db[d->player].location) ||
       (!could_doit(d->player, player, A_HLOCK) &&
        !controls(d->player, player, POW_WHO)))
      continue;
    strcpy(attr, atr_get(d->player, A_NOTIFY));
    if(!*attr)
      continue;

    if(!strcasecmp(attr, "all") || match_word(attr, db[player].name) ||
       (*(s=atr_get(player, A_ALIAS)) && match_word(attr, s)))
      notify(d->player, message);
  }
}

void display_cnotify(dbref player)
{
  DESC *d;
  char buf[9000], attr[8192], *s;

  strcpy(attr, atr_get(player, A_NOTIFY));
  if(!*attr)
    return;
  *buf='\0';

  for(d=descriptor_list;d;d=d->next) {
    if(!(d->flags & C_CONNECTED) || player == d->player ||
       (!could_doit(player, d->player, A_HLOCK) && !controls(player,
       d->player, POW_WHO)))
      continue;

    if(match_word(attr, db[d->player].name) || (*(s=atr_get(d->player,A_ALIAS))
       && match_word(attr, s)))
      strcat(buf, tprintf(" %s", db[d->player].name));
    if(strlen(buf) > 8000)
      break;
  }

  if(*buf)
    notify(player, tprintf("\e[1mConnected:\e[m %s", buf+1));
}

/* direct public speech to player in room */
void do_to(player,arg1,arg2,argv,pure)
dbref player;
char *arg1, *arg2, **argv, *pure;
{
  dbref thing;
  char *s;

  if(!pass_slock(player, 1))
    return;
  if(!(s=strchr(pure, ' '))) {
    notify(player, "No message.");
    return;
  }

  *s++='\0';
  if(!*pure) {
    notify(player, "No player mentioned.");
    return;
  } 
  if((thing=lookup_player(pure)) == AMBIGUOUS)
    thing=player;

  notify_except(db[player].location, NOTHING, tprintf(
       "%s [to %s]%s%s%s %s", db[player].name, thing==NOTHING ? pure:
       db[thing].name, (*s == ';')?": ":"", (*s == ';')?db[player].name:"",
       (*s == ':')?"":((*s == ';')?possname(db[player].name):":"),
       (*s == ':' || *s == ';') ? s+1 : s));
}

void do_talk(player, arg1)
dbref player;
char *arg1;
{
  dbref thing;
  char buf[528];

  if((thing=match_thing(player, arg1, MATCH_CON)) == NOTHING)
    return;
  if(Typeof(thing) != TYPE_THING) {
    notify(player, "This command works on non-player characters only.");
    return;
  } if(!*atr_get(thing, A_TALK)) {
    notify(player, "You can't talk to that.");
    return;
  }

  sprintf(buf, "talks to %s.", db[thing].name);
  did_it(player, thing, A_TALK, NULL, A_OTALK, buf, A_ATALK);
}
