/* db/save.c */

#include "externs.h"

#define DB_VERSION 1	// TinyMare Compressed Ram v1

extern int db_flags;


void putref(FILE *f, dbref num)
{
  if(num == HOME)
    num=db_top;
  else if(num < 0) {
    fputc(0xFF, f);
    return;
  }

  if(dbref_len >= 2) {
    if(dbref_len >= 3) {
      if(dbref_len == 4)
        fputc(num >> 24, f);
      fputc(num >> 16, f);
    }
    fputc(num >> 8, f);
  }
  fputc(num, f);
}

void putstring(FILE *f, char *s)
{
  if(s)
    fputs(s, f);
  fputc(0, f);
}


dbref db_write(FILE *f, char *filename)
{
  long pos, pos2;
  int i;
  
  /* determine database modifier flags */
  db_flags=0;

  if((unsigned long)time(NULL) > 0xFFFFFFFF)
    db_flags |= DB_LONGINT;  /* time must be stored using 64-bits */

  /* write database header */
  fprintf(f, "MARE");
  putchr(f, DB_VERSION);
  putnum(f, db_top);
  putnum(f, db_flags);
  putchr(f, NUM_POWS);

  /* select database reference number byte-length */
  dbref_len=(db_top < 0xFF)?1:(db_top < 0xFF00)?2:(db_top < 0xFF0000)?3:4;

  /* save configuration */
  for(i=0;db_rule[i].type;i++) {
    /* config header */
    putchr(f, db_rule[i].type);

    pos=ftell(f);
    db_rule[i].save(f, db_rule[i].type);

    /* check if any data has been written */
    pos2=ftell(f);
    if(pos2-pos < 1)
      fseek(f, -1, SEEK_CUR);  /* No configuration data, back up 1 byte */
  }
  fputc(0, f);

  /* write database */
  for(i=0;i<db_top;i++) {
    if(Typeof(i) == TYPE_GARBAGE)
      continue;
    db_write_object(f, i);
  }

  /* write closing -1 as end of list marker. check for out-of-disk errors. */
  if(fputc(0xFF, f) == EOF) {
    perror(filename);
    return 0;
  }

  return db_top;
}

void db_write_object(FILE *f, dbref i)
{
  ALIST *attr;
  
  /* save object header */
  putref(f, i);
  putstring(f, db[i].name);
  putref(f, db[i].location);
  putref(f, db[i].contents);
  putref(f, db[i].exits);
  putref(f, db[i].link);
  putref(f, db[i].next);
  putref(f, db[i].owner);
  putnum(f, db[i].flags);
//putnum(f, db[i].plane);
  putnum(f, db[i].pennies);

  if(db_flags & DB_LONGINT) {
    putlong(f, db[i].create_time);
    putlong(f, db[i].mod_time);
  } else {
    putnum(f, db[i].create_time);
    putnum(f, db[i].mod_time);
  }

  /* save object lists */
  putlist(f, db[i].parents);
  putlist(f, db[i].children);
  putlist(f, db[i].relay);
  putlist(f, db[i].zone);

  /* write out special player information */
  if(Typeof(i) == TYPE_PLAYER) {
    fputc(db[i].data->class-1, f);
    fputc(db[i].data->rows, f);
    fputc(db[i].data->cols, f);
    fputc((db[i].data->tzdst << 6) + ((db[i].data->tz & 0x1F) << 1) +
          db[i].data->gender, f);
    putshort(f, db[i].data->term);
    putnum(f, db[i].data->steps);
    putnum(f, db[i].data->sessions);
    putnum(f, db[i].data->age);

    if(db_flags & DB_LONGINT) {
      putlong(f, db[i].data->last);
      putlong(f, db[i].data->lastoff);
    } else {
      putnum(f, db[i].data->last);
      putnum(f, db[i].data->lastoff);
    }

    fwrite(db[i].data->pass, (db[i].data->term & TERM_MD5CRYPT)?20:10, 1, f);
    fwrite(db[i].data->powers, (NUM_POWS+3)/4, 1, f);
  }

  putshort(f, 0);
  putchr(f, 0);

  /* write the attribute list */
  for(attr=db[i].attrs;attr;attr=attr->next) {
    fputc(attr->num, f);
    putref(f, attr->obj);
    putstring(f, attr->text);
  }
  fputc(0, f);

  put_atrdefs(f, i);
}

void putlist(FILE *f, dbref *list)
{
  int a=0, *ptr;

  /* save list count */
  for(ptr=list;ptr && *ptr != NOTHING;ptr++,a++);
  putref(f, a-1);

  for(ptr=list;ptr && *ptr != NOTHING;ptr++)
    putref(f, *ptr);
}

void put_atrdefs(FILE *f, dbref thing)
{
  ATRDEF *k;

  for(k=db[thing].atrdefs;k;k=k->next) {
    putchr(f, k->atr.num);
    putshort(f, k->atr.flags);
    putstring(f, k->atr.name);
  }
  fputc(0, f);
}
