/* game/queue.c */
/* command queue processing */

#include "externs.h"

int queue_size[NUM_QUEUES];

static struct bque *qfirst, *qlast;
static struct bque *qwait, *qwaitlast;

long next_wait;


void immediate_que(dbref player, char *cmd, dbref thing)
{
  char buf[8192], buff[strlen(cmd)+1], *r, *s=buff;
  char *sptr[10];
  int a;

  /* check that the player can do a queue command */
  if((db[player].flags & (GOING|HAVEN)) || Typeof(player) == TYPE_GARBAGE)
    return;

  /* preserve local registers */
  for(a=0;a<10;a++)
    SPTR(sptr[a], wptr[a]);

  strcpy(buff, cmd);
  while((r=ansi_parse_up(&s, ';'))) {
    func_zerolev();
    pronoun_substitute(buf, player, thing, r);
    if(*buf)
      process_command(player, buf, thing, 0);

    /* restore local registers after each command */
    for(a=0;a<10;a++)
      strcpy(wptr[a], sptr[a]);
  }

  for(a=0;a<10;a++)
    free(sptr[a]);
}

void parse_que(dbref player, char *text, dbref cause)
{
  struct bque *new;
  char buff[strlen(text)+1], *cmd, *s=buff;
  int a=0;

  if((db[player].flags & (GOING|HAVEN)) || Typeof(player) == TYPE_GARBAGE)
    return;

  strcpy(buff, text);

  /* add commands to queue */
  while((cmd=ansi_parse_up(&s, ';'))) {
    /* make sure player can afford to do it */
    if(!power(player, POW_QUEUE)) {
      if(db[player].queue >= QUEUE_QUOTA) {
        db[player].flags |= HAVEN;
        notify(db[player].owner, tprintf("Run away object: %s",
               unparse_object(db[player].owner, player)));
        return;
      } if(!db[db[player].owner].pennies &&
           !charge_money(player, rand_num(QUEUE_COST)?0:1, 0)) {
        notify(db[player].owner, tprintf(
               "Not enough money to queue command on %s.",
               unparse_object(db[player].owner, player)));
        return;
      }
    }

    /* create pointer reference */
    new=malloc(sizeof(struct bque)+strlen(cmd)+1);
    new->next=NULL;
    new->player=player;
    new->cause=cause;
    for(a=0;a<10;a++) {
      if(*wptr[a])
        SPTR(new->env[a], wptr[a]);
      else
        new->env[a]=NULL;
    }
    strcpy(new->cmd, cmd);

    /* add to the queue list */
    if(qlast) {
      qlast->next=new;
      qlast=new; 
    } else
      qlast=qfirst=new;

    db[player].queue++;
    queue_size[Q_COMMAND]++;
  }
}

void do_wait(player, arg1, command, argv, pure, cause)
dbref player, cause;
char *arg1, *command, **argv, *pure;
{
  struct bque *new;
  int a, wait=atoi(arg1), pow_queue=power(player, POW_QUEUE);

  if(db[player].flags & (GOING|HAVEN))
    return;

  /* execute command immediately if no time is specified */
  if(wait < 1) {
    parse_que(player, command, cause);
    return;
  } else if(wait > 31536000)  /* 1 year maximum */
    wait=31536000;

  /* make sure player can afford to do it */
  if(!pow_queue) {
    if(db[player].queue >= QUEUE_QUOTA) {
      db[player].flags |= HAVEN;
      notify(db[player].owner, tprintf("Run away object: %s",
             unparse_object(db[player].owner, player)));
      return;
    } if(!db[db[player].owner].pennies) {
      notify(db[player].owner,tprintf("Not enough money to queue @wait on %s.",
             unparse_object(db[player].owner, player)));
      return;
    }
  }

  if(queue_size[Q_WAIT] >= (pow_queue?MAX_WIZARD_WAIT:MAX_PLAYER_WAIT)) {
    notify(db[player].owner,tprintf("%s: Too many @wait commands in progress.",
           unparse_object(db[player].owner, player)));
    return;
  }

  /* create pointer reference */
  new=malloc(sizeof(struct bque)+strlen(command)+1);
  new->next=NULL;
  new->player=player;
  new->cause=cause;
  new->wait=now+wait;
  for(a=0;a<10;a++) {
    if(*wptr[a])
      SPTR(new->env[a], wptr[a]);
    else
      new->env[a]=NULL;
  }
  strcpy(new->cmd, command);

  /* add to the @wait queue */
  if(qwaitlast) {
    qwaitlast->next=new;
    qwaitlast=new; 
  } else
    qwaitlast=qwait=new;

  /* new value for @wait queue to wake up on */
  if(!next_wait || new->wait < next_wait)
    next_wait=new->wait;

  db[player].queue++;
  queue_size[Q_WAIT]++;
}

/* execute one round of queue commands */
void flush_queue()
{
  struct bque *ptr;
  int a, i, num=queue_size[Q_COMMAND];
  char buff[8192];

  if(!num)
    return;

  nstat[NS_QUEUE]++;
  for(i=0;i<num && qfirst;i++) {
    /* check time slice */
    if(game_online && get_time() > time_slice) {
      nstat[NS_OVERRUNS]++;
      break;
    }

    ptr=qfirst;
    if(!(qfirst=ptr->next))
      qlast=NULL;
    if(ptr->player != NOTHING) {
      db[ptr->player].queue--;
      if(!(db[ptr->player].flags & (GOING|HAVEN))) {
        for(a=0;a<10;a++)
          strcpy(wptr[a], ptr->env[a]?:"");
        func_zerolev();
        pronoun_substitute(buff, ptr->player, ptr->cause, ptr->cmd);
        if(*buff)
          process_command(ptr->player, buff, ptr->cause, 0);
      }
    }

    /* remove pointer */
    for(a=0;a<10;a++)
      free(ptr->env[a]);
    free(ptr);

    queue_size[Q_COMMAND]--;
  }

  for(a=0;a<10;a++)
    *wptr[a]='\0';
}

/* called every second to check for wait queue commands */           
void wait_second()
{
  struct bque *ptr, *next, *last=NULL;
  int a;

  if(!next_wait || next_wait > now)  /* nothing to do yet */
    return;
  next_wait=0;

  for(ptr=qwait;ptr;ptr=next) {
    next=ptr->next;
    if(ptr->wait > now && !(db[ptr->player].flags & (GOING|HAVEN))) {
      last=ptr;
      if(!next_wait || ptr->wait < next_wait)
        next_wait=ptr->wait;
      continue;
    }

    /* process @wait command */
    db[ptr->player].queue--;
    queue_size[Q_WAIT]--;
    if(!(db[ptr->player].flags & (GOING|HAVEN))) {
      for(a=0;a<10;a++)
        strcpy(wptr[a], ptr->env[a]?:"");
      parse_que(ptr->player, ptr->cmd, ptr->cause);
    }

    /* remove from queue */
    if(last)
      last->next=ptr->next;
    else
      qwait=ptr->next;
    for(a=0;a<10;a++)
      free(ptr->env[a]);
    free(ptr);
  }
  qwaitlast=last;
}

/* tell player what commands they have pending in the queue */
void do_queue(dbref player, char *arg1, char *arg2)
{
  const char *options[]={"All", "Immediate", "Waited", 0};

  struct bque *tmp;
  dbref thing=power(player, POW_QUEUE)?AMBIGUOUS:player;
  int type=0, count, tcount=0;
  
  /* Select a type of viewing */
  if(*arg1) {
    for(type=0;options[type];type++)
      if(string_prefix(options[type], arg1))
        break;
    if(!options[type]) {
      notify(player, tprintf("No such queue '%s'.", arg1));
      return;
    }
  }

  /* If present, only view the queue on <arg2> */
  if(*arg2) {
    if((thing=match_thing(player, arg2, MATCH_EVERYTHING)) == NOTHING)
      return;
    if(!controls(player, thing, POW_QUEUE)) {
      notify(player, "Permission denied.");
      return;
    }
  }

  /* Immediate commands */
  if(type < 2) {
    for(count=0,tmp=qfirst;tmp;tmp=tmp->next)
      if(!Invalid(tmp->player) && (thing == AMBIGUOUS ||
         thing == tmp->player || thing == db[tmp->player].owner) &&
         controls(player, tmp->player, POW_QUEUE)) {
        if(!count)
          notify(player, "\e[1mImmediate commands:\e[m");
        notify(player, tprintf("%s:%s", unparse_object(player, tmp->player),
               tmp->cmd));
        count++, tcount++;
      }
    if(count > 0)
      notify(player, tprintf("-- Total: %d Immediate --", count));
  }

  /* Waited Commands */
  if(!type || type == 2) {
    for(count=0,tmp=qwait;tmp;tmp=tmp->next)
      if(!Invalid(tmp->player) && (thing == AMBIGUOUS ||
         thing == tmp->player || thing == db[tmp->player].owner) &&
         controls(player, tmp->player, POW_QUEUE)) {
        if(!count)
          notify(player, "\e[1m@waited commands:\e[m");
        notify(player, tprintf("\e[1;36m[\e[37m%ld\e[36m]\e[m%s:%s", 
               tmp->wait-now, unparse_object(player, tmp->player), tmp->cmd));
        count++, tcount++;
      }
    if(count > 0)
      notify(player, tprintf("-- Total: %d @waited --", count));
  }

  if(!tcount)
    notify(player, "Nothing in the queue.");
}

/* remove all queued commands from a certain player */
void do_halt(player, arg1, ncom, argv, pure, cause)
dbref player, cause;
char *arg1, *ncom, **argv, *pure;
{
  int thing=player;

  if(!strcmp(arg1, "ALL") && power(player, POW_SECURITY))
    thing=AMBIGUOUS;
  else if(*arg1) {
    if((thing=match_thing(player, arg1, MATCH_EVERYTHING)) == NOTHING)
      return;
    if(!controls(player, thing, POW_MODIFY)) {
      notify(player, "Permission denied.");
      return;
    }
  }

  if(!Quiet(player)) {
    if(thing == player)
      notify(player, "Everything halted.");
    else if(thing == AMBIGUOUS)
      notify(player, "Halted away every blessed thing.");
    else
      notify(player, tprintf("%s halted.", unparse_object(player, thing)));
  }

  /* remove the queue commands */
  halt_object(thing);

  /* execute extra command? */
  if(*ncom && !Invalid(thing))
    parse_que(thing,ncom,cause);
}

void halt_object(dbref thing)
{
  struct bque *ptr, *next, *last=NULL;
  int a;

  /* remove regular queue commands */
  for(ptr=qfirst;ptr;ptr=ptr->next)
    if(ptr->player != NOTHING && (thing == AMBIGUOUS || ptr->player == thing ||
       db[ptr->player].owner == thing)) {
      db[ptr->player].queue--;
      ptr->player=NOTHING;
    }

  /* remove wait queue commands */
  for(ptr=qwait;ptr;ptr=next) {
    next=ptr->next;
    if(ptr->player != thing && db[ptr->player].owner != thing &&
       thing != AMBIGUOUS) {
      last=ptr;
      continue;
    }

    db[ptr->player].queue--;
    queue_size[Q_WAIT]--;

    /* remove from queue */
    if(last)
      last->next=ptr->next;
    else
      qwait=ptr->next;
    for(a=0;a<10;a++)
      free(ptr->env[a]);
    free(ptr);
  }
  qwaitlast=last;
}
